(*
    Copyright (C) 2012  G. Gonthier, B. Ziliani, A. Nanevski, D. Dreyer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

Require Import Coq.Lists.List.
Set Implicit Arguments.
Unset Strict Implicit.

Obligation Tactic := idtac.

(******************************************************************************)
(*       Module for searching and inserting elements in a list                *)
(******************************************************************************)

Class XFind A (s : list A) (e : A) := {
  seq_of : list A;
  index_of : nat;
  xfind : nth index_of seq_of e = e}.

Implicit Arguments XFind [A].

Program Instance found_struct A (x:A) t : XFind (x :: t) x := 
  {| seq_of := (x :: t); index_of := 0|}.
Next Obligation.
auto.
Qed.

Program Instance recurse_struct A (y:A) t e (f : XFind t e) :
  XFind (y :: t) e | 100 := {| seq_of := (y :: seq_of); index_of := (S index_of)|}.
Next Obligation.
intros; case f; auto.
Qed.

Program Instance extend_struct A (x:A) : XFind nil x := 
  {| seq_of := (x :: nil); index_of := 0|}.
Next Obligation.
auto.
Qed.

Example unit_test A (x1 x2 x3 x y : A):
   (forall s e (f : XFind s e), nth index_of seq_of x1 = e -> e = x) ->
  x = x.
Proof.
intro test_form.
eapply (test_form nil). simpl.
eapply (test_form (x1 :: x :: nil)). simpl.
try eapply (test_form (x1 :: x2 :: x :: x3 :: nil)). simpl.
try eapply (test_form (x1 :: x2 :: x3 :: nil)). simpl.
Abort.

