Require Import mtac.
Require Import List.

Set Implicit Arguments.
Unset Strict Implicit.


Definition NotFound : Exception.  exact exception.  Qed.

Notation "<| p |>" := p%mtac_patt.

Unset Use Munify.

Set Printing Existential Instances.

Definition CaseT := forall A x 
  (f : forall s' : list A, M ((fun s : list A => In x s) s')) s, 
  tpatt (list A) (fun s' : list A => In x s') s.

Class Case (n : nat) := { the_case : CaseT }.

Definition build_cases : M (list CaseT) := 
  let cases :=
    mfix f (n : nat) (l : list CaseT) : M (list CaseT) :=
      X <- evar (Case n);
      solve_typeclasses;;
      b <- is_evar X;
      if b then 
        ret l
      else
        f (S n) (@the_case _ X :: l)
  in cases 0 nil.

Definition last_case : M nat :=
  l <- build_cases;
  retS (length l).

Definition default_case : CaseT := 
  fun A x f s => <| [l] l =m> raise NotFound |>.

Instance default_inst : Case 0 := {| the_case := default_case |}.

Definition iterate_case : CaseT := 
  fun A x f s => <| [y s'] (y :: s') =m> r <- f s'; ret (in_cons y _ _ r) |>.

Definition iterate_inst' : Case (run last_case) := {| the_case := iterate_case |}.
Instance iterate_inst : Case _ := iterate_inst'.

Definition found_case : CaseT :=
  fun A x f s => <| [s'] (x :: s') =m> ret (in_eq _ _) |>.

Definition found_inst' : Case (run last_case)  := {| the_case := found_case |}.
Instance found_inst : Case _ := found_inst'.  


Definition inlist_ext {A} (x : A) s :=
  cs <- build_cases;
  cs <- retS (map (fun (c : CaseT) => c A x) cs);
  (mfix1 f (s : list A) : M (In x s) := 
    csap <- retS (map (fun c => c f s) cs);
    tmatch (fun s'=>In x s') s csap) s.



Definition app_case : CaseT := fun A x f s => 
  <| [ l r ] l ++ r =m>
      mtry 
        il <- f l;
        ret (in_or_app l r x (or_introl il))
      with NotFound =>
        ir <- f r;
        ret (in_or_app l r x (or_intror ir))
      end |>.



Import ListNotations.

Check (run (inlist_ext 0 [1;0])).

Example ex_no_app_l (x1 x2 : nat) s : In x2 ([x1; x2] ++ s) := run (inlist_ext _ _).

Fail Example ex_no_app_r (x1 x2 : nat) s : In x1 (s ++ [x1; x2]) := run (inlist_ext _ _).

Definition app_inst' : Case (run last_case)  := {| the_case := app_case |}.
Instance app_inst : Case _ := app_inst'.  

Example ex_app_r (x1 x2 : nat) s : In x1 (s ++ [x1; x2]) := run (inlist_ext _ _).
