Require Import mtac.
Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Add LoadPath "../../examples-lo/".
Require Import heaps.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Lemma same_ptr (x : ptr) (A : Type) (v : A) : 
  def (x :-> v) -> x \in dom (x :-> v).
Proof.
by rewrite defPt domPt inE /= eq_refl. 
Qed.

Lemma inLeft (x : ptr) (l r : heap) :
   (def l -> x \in dom l) -> def (l :+ r) -> x \in dom (l :+ r).
Proof.
move=>H D.
by rewrite domUn !inE /= D (H (defUnl D)).
Qed.

Lemma inRight (x : ptr) (l r : heap) :
   (def r -> x \in dom r) -> def (l :+ r) -> x \in dom (l :+ r).
Proof.
move=>H D.
by rewrite domUn !inE /= D (H (defUnr D)) orbT.
Qed.

Definition NotFound : Exception. exact exception. Qed.

(** Mtactic to search a pointer [x] in heap [h] *)
Program Definition indomP (x : ptr)  :=
  mfix f (h : heap) : M (def h -> x \in dom h) :=
    mmatch h as h' return M (def h' -> x \in dom h') with
      | [A (v:A)] x :-> v => ret (@same_ptr x _ v)
      | [l r] l :+ r => 
        mtry 
          a <- f l;
          ret (inLeft a)
        with _ =>
          a <- f r;
          ret (inRight a)
        end
      | _ => raise NotFound
    end.

Definition indom x h (D : def h) := (r <- indomP x h; retS (r D)).

Implicit Arguments indom [x h].

(*************************************************)
(*                   Examples                    *)
(*************************************************)

Example appl0 A (x1 x2 : ptr)
              (v1 v2  : A) :
          def (x1 :-> v1 :+ x2 :-> v2) -> 
         x2 \in dom (x1 :-> v1 :+ x2 :-> v2).
Proof.
move=>D.
Time by rewrite (eval (indom _)).
Time Qed.

Example appl1 A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h00 h01 h02 h03 h04 h05 h06 h07 h08 h09 : heap)
              (h10 h11 h12 h13 h14 h15 h16 h17 h18 h19 : heap)
              (h20 h21 h22 h23 h24 h25 h26 h27 h28 h29 : heap)
              (h30 h31 h32 h33 h34 h35 h36 h37 h38 h39 : heap)
              (h40 h41 h42 h43 h44 h45 h46 h47 h48 h49 : heap) :
          def (h01 :+ h02 :+ h03 :+ h04 :+ h05 :+ h06 :+ h07 :+ h08 :+ h09 :+
               h41 :+ h42 :+ h43 :+ h44 :+ h45 :+ h46 :+ h47 :+ h48 :+ h49 :+
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               h11 :+ h12 :+ h13 :+ h14 :+ h15 :+ h16 :+ h17 :+ h18 :+ h19 :+
               h21 :+ h22 :+ h23 :+ h24 :+ h25 :+ h26 :+ h27 :+ h28 :+ h29 :+
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               h31 :+ h32 :+ h33 :+ h34 :+ h35 :+ h36 :+ h37 :+ h38 :+ h39) -> 
         x7 \in dom (h01 :+ h02 :+ h03 :+ h04 :+ h05 :+ h06 :+ h07 :+ h08 :+ h09 :+
               h41 :+ h42 :+ h43 :+ h44 :+ h45 :+ h46 :+ h47 :+ h48 :+ h49 :+
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               h11 :+ h12 :+ h13 :+ h14 :+ h15 :+ h16 :+ h17 :+ h18 :+ h19 :+
               h21 :+ h22 :+ h23 :+ h24 :+ h25 :+ h26 :+ h27 :+ h28 :+ h29 :+
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               h31 :+ h32 :+ h33 :+ h34 :+ h35 :+ h36 :+ h37 :+ h38 :+ h39).
Proof.
move=>D.
Time by apply: (run (indom _)).
Time Qed.


(* simple example *)
Example ex1 A (x1 x2 : ptr) (v1 v2 : A) (h1 h2 : heap) :
          def (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) -> 
          if x2 \in dom (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) return bool
            then 1 == 1 
            else 1 == 0.
Proof.
move=>D.
by rewrite (eval (indom _)).
Qed.
