Set Automatic Coercions Import.

Require Import mtac.
Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Add LoadPath "../../examples-lo/".
Require Import prefix.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(******************************************************************************)
(*       Module for searching and inserting elements in a list                *)
(******************************************************************************)

Section XFind.

Variable A : Type.

Definition invariant s r i (e : A) := onth r i = Some e /\ prefix s r.

(* Main structure
   s : input sequence
   r : output sequence. If elem_of is in the sequence, then it's equal to s, 
       otherwise it's equal to (elem_of :: s)
   i : output index of elem_of in r *)
Structure xfind e s := XFind {
  seq_of : seq A;
  index_of : nat;
  inv_of : invariant s seq_of index_of e}.

Implicit Arguments XFind [].

Lemma found_pf x t : invariant (x :: t) (x :: t) 0 x. 
Proof. by split; [|apply: prefix_refl]. Qed.

Lemma extend_pf x : invariant [::] [:: x] 0 x.
Proof. by []. Qed.

Lemma recurse_pf e (y : A) (s : seq A) (f : xfind e s) :
        invariant (y :: s) (y :: (seq_of f)) (index_of f).+1 e.
Proof. by case:f=>[r i [H1 H2]]; split; [|apply/prefix_cons]. Qed.

Set Printing Existential Instances.

Definition NotUnifiableWithSeq : Exception. exact exception. Qed.

Definition findx (e : A) :=
  mfix fx (s : seq A) : M (xfind e s) :=
  mmatch s with
  | ([::]) => ret (XFind e [::] [:: e] 0 (extend_pf e))
  | [s'] (e :: s') => ret (XFind e (e :: s') (e :: s') 0 (found_pf e s'))
  | [y s'] y :: s' => 
      f <- fx s';
      ret (XFind e (y :: s') (y :: seq_of f) (index_of f).+1 (recurse_pf y f)) 
  | [s'] s' => raise NotUnifiableWithSeq
  end.
 

End XFind.
